#include <os.h>
#include "gfx.h"
#include "sprites.h"

#define RAM_PALETTE_ADDRESS 0xC0000200

inline void setPixelBuf(uint8_t* scrbuf, unsigned short x, unsigned short y, uint8_t color){
	uint8_t* p = scrbuf + ((x >> 1) + (y << 7) + (y << 5));
	if(x < SCREEN_WIDTH && y < SCREEN_HEIGHT)
		*p = (x & 1) ? ((*p & 0xF0) | color) : ((*p & 0x0F) | (color << 4));
}

void putSprite(uint8_t* scrbuf, uint8_t spritenum, uint16_t x, uint16_t y, uint8_t color){
    uint8_t w,h;
	for(h=0;h<7;h++){
	    for(w=7;w>0;w--){
	        if(sprites[spritenum][h] & (0x1 << (7-w)))
                setPixelBuf(scrbuf,x+w-1,y+h,color);
            else
                setPixelBuf(scrbuf,x+w-1,y+h,0x0);
	    }
	}
}


/*
0    000000     0000    schwarz
((B    FFB897	    7EF2    pink:dots))
2    4A4A00     2520    dunkle tarnfarbe
3    2121DE	    109B    blau:wnde&augen
4    FF0000	    7C00    rot
5    008080	    0210    dunkelcyan:melone
6    9C5A3C     4D67    braun:kirschenstiel&andereboni
7    00FFDE	    03FB    trkis
8    949400     4A40    tarnfarbe
9    00FF00	    03E0    grn:melone&andereboni
A    DE9747	    6E48    braun:kirschenstiel&andereboni
B    FFB847	    7EE8    orange:gegner&pfirsich
C    FFB8DE	    7EFB    pink:pinky&ausgang
D    DEDEDE	    6F7B    weissgrau:schrift&glanzaufkirschen&augen
E    FFFF00	    7FE0    gelb:schrift&pacman
*/


void initPalette(){
    const int palette[8] = {//little endian entry order, 31 unused, 30 - 26 r, 25 - 21 g, 20 - 16 b, 15 unused, 14 - 10 r, 9 - 5 g, 4 - 0 b
        0x7EF20000,
        0x109B2520,
        0x02107C00,
        0x03FB4D67,
        0x03E04A40,
        0x7EE86E48,
        0x6F7B7EFB,
        0xFFFF7FE0
    };
    *(uint32_t*)0xC0000018 = (*(uint32_t*)0xC0000018 & 0x0000FFF1) | 0x00000404;//4 bit per pixel, big endian pixel order
    memcpy((uint32_t*)RAM_PALETTE_ADDRESS,palette,32);//32 bytes = 8 words = 16 colors/palette entries
}

void drawTest(uint8_t* scrbuf){
    uint8_t color, i, y;
    for(color=0x0;color<0x10;color++){
        for(i=0;i<20;i++){
            for(y=0;y<20;y++)
                setPixelBuf(scrbuf,color*20+i,y,color);
        }
    }
}
