#include <os.h>
#include "level.h"
#include "pacmansprites.h"
#include "enemysprites.h"
#include "gfx.h"

#define SCREEN_POINTER_ADDRESS 0xC0000010

#define LEVEL_X_OFFSET ((SCREEN_WIDTH - 28*7)/2)
#define LEVEL_Y_OFFSET 13

#define DIR_DOWN 0x01
#define DIR_UP 0x02
#define DIR_RIGHT 0x04
#define DIR_LEFT 0x08

//all coordinates of pacman and enemies are relative to the upper right corner of the board, i. e. from 0 to levelwidth*tilewidth - 1
//to do:    -screen off while changing mode
//          -cornering

static uint8_t distance(uint16_t x1, uint16_t y1, uint16_t x2, uint16_t y2){
    //return isqrt((x1-x2)*(x1-x2) + (y1-y2)*(y1-y2));
    return abs(x1-x2) + abs(y1-y2);//here it isn't necessary to calculate the geometric distance
}

char newdir(char enemydir, uint16_t enemyx, uint16_t enemyy, uint16_t targetx, uint16_t targety){//finds the direction the enemy will take after the next tile
    uint8_t tmpdist = 0;
    char tmpdir;
    uint8_t tmp;
    if(enemydir == DIR_UP)
        enemyy-=1;
    else if(enemydir == DIR_LEFT)
        enemyx-=1;
    else if(enemydir == DIR_DOWN)
        enemyy+=1;
    else if(enemydir == DIR_RIGHT)
        enemyx+=1;
    if(enemyx == 0 || enemyx == 27)
        return enemydir;
    if(enemydir != DIR_DOWN && walls[enemyy-1][enemyx] < 3){
        tmpdist = distance(enemyx, enemyy-1, targetx, targety);
        tmpdir = DIR_UP;
    }
    if(enemydir != DIR_RIGHT && walls[enemyy][enemyx-1] < 3){
        if(!tmpdist || (tmp = distance(enemyx-1, enemyy, targetx, targety)) < tmpdist){
            tmpdist = tmp;
            tmpdir = DIR_LEFT;
        }
    }
    if(enemydir != DIR_UP && walls[enemyy+1][enemyx] < 3){
        if(!tmpdist || (tmp = distance(enemyx,enemyy+1,targetx,targety)) < tmpdist){
            tmpdist = tmp;
            tmpdir = DIR_DOWN;
        }
    }
    if(enemydir != DIR_LEFT && walls[enemyy][enemyx+1] < 3){
        if(!tmpdist || (tmp = distance(enemyx+1,enemyy,targetx,targety)) < tmpdist){
            tmpdist = tmp;
            tmpdir = DIR_RIGHT;
        }
    }
    return tmpdir;
}

void drawPacman(uint8_t* scrbuf, char pacmandir, uint16_t animx, uint16_t animy){
    static uint8_t frame = 0;
    uint8_t color;
    uint8_t x,y;
    uint8_t (*sprite)[6];

    if(pacmandir == 0){
        sprite = pacman;
    }
    else if(pacmandir == DIR_RIGHT){
        sprite = pacman_right[frame/2];
    }
    else if(pacmandir == DIR_LEFT){
        sprite = pacman_left[frame/2];
    }
    else if(pacmandir == DIR_DOWN){
        sprite = pacman_down[frame/2];
    }
    else if(pacmandir == DIR_UP){
        sprite = pacman_up[frame/2];
    }
    else{
        return;
    }

    for(y=0;y<11;y++){
        for(x=0;x<11;x++){
            color = x%2 ? sprite[y][x/2]&0x0F : sprite[y][x/2]>>4;
            if(color)
                setPixelBuf(scrbuf, LEVEL_X_OFFSET+animx+x - (11/2), LEVEL_Y_OFFSET+animy+y - (11 / 2), color);
        }
    }

    frame++;
    if(frame > 3)
        frame = 0;
}

void drawWalls(uint8_t* levelbuf){
    //memset(levelbuf,0x00,SCREEN_BYTES_SIZE);
    uint8_t x,y;
    for(y=0;y<31;y++){
        for(x=0;x<28;x++){
            if(walls[y][x] == 1 || walls[y][x] == 2)
                putSprite(levelbuf, walls[y][x], x*7, y*7, 0xB);
            else if(walls[y][x] == 25)
                putSprite(levelbuf, walls[y][x], x*7, y*7, 0xC);
            else
                putSprite(levelbuf, walls[y][x], x*7, y*7, 0x3);
        }
    }
}

uint32_t frightenedtimeelapsed(uint32_t start){
    static uint32_t starttime;
    if(start)
        starttime = start;
    return *(uint32_t*)0x90090000 - starttime;
}

void drawEnemies(uint8_t* scrbuf, uint16_t* enemyx, uint16_t* enemyy, uint8_t* mode, uint8_t frighttime){
    uint8_t i,x,y,color;
    for(i=0;i<4;i++){
        if(mode[i] < 2){
            for(y=0;y<11;y++){
                for(x=0;x<11;x++){
                    color = x%2 ? enemies[i][y][x/2]&0x0F : enemies[i][y][x/2]>>4;
                    if(color)
                        setPixelBuf(scrbuf, LEVEL_X_OFFSET+(enemyx[i]>>8)+x - (11/2), LEVEL_Y_OFFSET+(enemyy[i]>>8)+y - (11/2), color);
                }
            }
        }
        else if(frightenedtimeelapsed(0) < frighttime-2){
            for(y=0;y<11;y++){
                for(x=0;x<11;x++){
                    color = x%2 ? enemies[i][y][x/2]&0x0F : enemies[i][y][x/2]>>4;
                    if(color)
                        setPixelBuf(scrbuf, LEVEL_X_OFFSET+(enemyx[i]>>8)+x - (11/2), LEVEL_Y_OFFSET+(enemyy[i]>>8)+y - (11/2), 0x3);
                }
            }
        }
        else{
            for(y=0;y<11;y++){
                for(x=0;x<11;x++){
                    color = x%2 ? enemies[i][y][x/2]&0x0F : enemies[i][y][x/2]>>4;
                    if(color)
                        setPixelBuf(scrbuf, LEVEL_X_OFFSET+(enemyx[i]>>8)+x - (11/2), LEVEL_Y_OFFSET+(enemyy[i]>>8)+y - (11/2), 0x5);
                }
            }
        }
    }
}

void copyLevelbuf(uint8_t* scrbuf, uint8_t* levelbuf){
    int y;
    for(y=0;y<SCREEN_HEIGHT;y++)
        memcpy((scrbuf + ((LEVEL_Y_OFFSET + y)*SCREEN_WIDTH/2) + LEVEL_X_OFFSET/2), (levelbuf + (y*SCREEN_WIDTH/2)), (28*7/2));
}

uint32_t elapsedtime(uint32_t start){
    static uint32_t starttime;
    if(start)
        starttime = start;
    return *(uint32_t*)0x90090000 - starttime;
}

void drawLifes(uint8_t* scrbuf, uint8_t lifes){
    uint8_t i;
    for(i=0;i<lifes;i++){
        putSprite(scrbuf,26,LEVEL_X_OFFSET+i*14+4,LEVEL_Y_OFFSET+31*7,0xE);
    }
}

void drawScore(uint8_t* scrbuf, unsigned score){
    char scorestring[11];
    sprintf(scorestring,"%i",score);
    uint8_t i;
    for(i=0;scorestring[i] != '\0'; i++){
        putSprite(scrbuf,scorestring[i]-21,LEVEL_X_OFFSET+4+i*7,0,0xD);
    }
}

void bugpatch(uint8_t* levelbuf){//quick fix since I don't know what causes the bug
    uint8_t i;
    for(i=0;i<28;i++)
        putSprite(levelbuf,walls[0][i],i*7,0,0x3);
    for(i=0;i<28;i++){
        if(walls[1][i] == 1)
            putSprite(levelbuf,walls[1][i],i*7,7,0xB);
        else
            putSprite(levelbuf,walls[1][i],i*7,7,0x3);
    }
}

int main(){
    if(is_cx){
        initPalette();
    }
    uint8_t* scrbuf = (uint8_t*)malloc(SCREEN_BYTES_SIZE);
    uint8_t* levelbuf = (uint8_t*)malloc(31*7*28*7);
    uint8_t* levelresetbuf = (uint8_t*)malloc(31*28);
    memset(scrbuf,0x00,SCREEN_BYTES_SIZE);
    memset(levelbuf,0x00,31*7*28*7);
    memcpy(levelresetbuf,&walls[0][0],31*28);
    if(scrbuf == NULL || levelbuf == NULL || levelresetbuf == NULL)
        return 1;

    int i;

    uint16_t pacmanx;//pacmanx = x-coordinate of pacman's center point
    uint16_t pacmany;
    uint16_t pacmanvel;
    char pacmandir;
    char pacmannextdir;

    uint16_t enemyx[4];
    uint16_t enemyy[4];
    uint16_t enemyvel[4];
    char enemydir[4];
    char enemynextdir[4];
    uint8_t enemylock[4];
    uint8_t mode[4];

    uint8_t eatendots;
    uint32_t score = 0;
    uint8_t lifes = 3;
    uint8_t level = 1;
    uint8_t frighttime;

    //drawTest(scrbuf);

    //int fps = 0;
    //uint32_t oldtime = 0;


    BEGINNING:

    pacmanx = (14*7) << 8;//pacmanx = x-coordinate of pacman's center point
    pacmany = (23*7 + 3) << 8;
    pacmandir = DIR_LEFT;
    pacmannextdir = pacmandir;

    enemyx[0] = (14*7)<<8;
    enemyx[1] = (12*7)<<8;
    enemyx[2] = (13*7+5)<<8;
    enemyx[3] = (15*7+3)<<8;
    enemyy[0] = (11*7+3)<<8;
    enemyy[1] = (14*7+3)<<8;
    enemyy[2] = (14*7+3)<<8;
    enemyy[3] = (14*7+3)<<8;
    enemydir[0] = DIR_LEFT;
    enemydir[1] = 0;
    enemydir[2] = 0;
    enemydir[3] = 0;
    enemynextdir[0] = newdir(enemydir[0],(enemyx[0]>>8)/7,(enemyy[0]>>8)/7,(pacmanx>>8)/7,(pacmany>>8)/7);
    enemynextdir[1] = 0;
    enemynextdir[2] = 0;
    enemynextdir[3] = 0;
    enemylock[0] = 0;
    enemylock[1] = 0;
    enemylock[2] = 0;
    enemylock[3] = 0;
    mode[0] = 0;
    mode[0] = 1;
    mode[0] = 2;
    mode[0] = 3;

    eatendots = 0;

    if(level == 1){
        enemyvel[0] = 114;//6fach zu anfang
        enemyvel[1] = 102;
        enemyvel[2] = 102;
        enemyvel[3] = 90;
        pacmanvel = 120;
        frighttime = 6;
    }
    else if(level < 5){
        memcpy(walls, levelresetbuf, 31*28);
        enemyvel[0] = 144;
        enemyvel[1] = 132;
        enemyvel[2] = 132;
        enemyvel[3] = 120;
        pacmanvel = 150;
        frighttime = 7 - level;
    }
    else if(level < 20){
        if(level == 5)
            frighttime = 2;
        else if(level == 6)
            frighttime = 5;
        else if(level == 7)
            frighttime = 2;
        else if(level == 8)
            frighttime = 2;
        else if(level == 9)
            frighttime = 1;
        else if(level == 10)
            frighttime = 5;
        else if(level == 11)
            frighttime = 2;
        else if(level == 12)
            frighttime = 1;
        else if(level == 13)
            frighttime = 1;
        else if(level == 14)
            frighttime = 3;
        else if(level == 15)
            frighttime = 1;
        else if(level == 16)
            frighttime = 1;
        else if(level == 17)
            frighttime = 0;
        else if(level == 18)
            frighttime = 1;
        else
            frighttime = 0;
        memcpy(walls, levelresetbuf, 31*28);
        enemyvel[0] = 162;
        enemyvel[1] = 156;
        enemyvel[2] = 156;
        enemyvel[3] = 150;
        pacmanvel = 168;
    }
    else{
        frighttime = 0;
        memcpy(walls, levelresetbuf, 31*28);
        enemyvel[0] = 162;
        enemyvel[1] = 156;
        enemyvel[2] = 156;
        enemyvel[3] = 150;
        pacmanvel = 150;
    }

    drawWalls(levelbuf);
    elapsedtime(*(uint32_t*)0x90090000);

    while(!isKeyPressed(KEY_NSPIRE_ESC) && !on_key_pressed()){

        /*if(oldtime != *(uint32_t*)0x90090000){
            oldtime = *(uint32_t*)0x90090000;
            printf("%i\t",fps);
            fps = 0;
        }
        ++fps;*/

        if(elapsedtime(0) == 4){
            enemyx[2] = (13*7+3)<<8;
            enemydir[2] = DIR_UP;
        }
        if(elapsedtime(0) == 8){
            enemyx[1] = (13*7+3)<<8;
            enemydir[1] = DIR_UP;
        }
        if(elapsedtime(0) == 10){
            enemyx[3] = (13*7+3)<<8;
            enemydir[3] = DIR_UP;
        }

        bugpatch(levelbuf);
        copyLevelbuf(scrbuf, levelbuf);
        drawPacman(scrbuf, pacmandir, pacmanx>>8, pacmany>>8);
        drawEnemies(scrbuf,enemyx,enemyy,mode,frighttime);
        drawLifes(scrbuf,lifes);
        drawScore(scrbuf,score);
        memcpy(*(uint8_t**)SCREEN_POINTER_ADDRESS,scrbuf,SCREEN_BYTES_SIZE);

        if(isKeyPressed(KEY_NSPIRE_UP))
            pacmannextdir = DIR_UP;
        else if(isKeyPressed(KEY_NSPIRE_LEFT))
            pacmannextdir = DIR_LEFT;
        else if(isKeyPressed(KEY_NSPIRE_RIGHT))
            pacmannextdir = DIR_RIGHT;
        else if(isKeyPressed(KEY_NSPIRE_DOWN))
            pacmannextdir = DIR_DOWN;

        for(i=0;i<4;i++){
            switch(enemydir[i]){
                case DIR_UP:
                    enemyy[i] -= enemyvel[i];
                break;
                case DIR_LEFT:
                    enemyx[i] -= enemyvel[i];
                break;
                case DIR_DOWN:
                    enemyy[i] += enemyvel[i];
                break;
                case DIR_RIGHT:
                    enemyx[i] += enemyvel[i];
                break;
            }
        }

        if(!(((pacmanx >> 8) - 3) % 7) && !(((pacmany >> 8) - 3) % 7)){
            if(walls[(pacmany >> 8) / 7][(pacmanx >> 8) / 7] == 1){
                walls[(pacmany >> 8) / 7][(pacmanx >> 8) / 7] = 0;
                eatendots++;
                score += 10;
                drawWalls(levelbuf);
            }
            else if(walls[(pacmany >> 8) / 7][(pacmanx >> 8) / 7] == 2){
                walls[(pacmany >> 8) / 7][(pacmanx >> 8) / 7] = 0;
                eatendots++;
                score += 50;
                memset(mode,2,4);
                frightenedtimeelapsed(*(uint32_t*)0x90090000);
                drawWalls(levelbuf);
            }

            if(eatendots == 244){
                level++;
                goto BEGINNING;
            }

            switch(pacmandir){
                case DIR_UP:
                    if(walls[((pacmany >> 8) / 7) - 1][(pacmanx >> 8) / 7] > 2)
                        pacmandir = 0;
                break;
                case DIR_LEFT:
                    if(walls[(pacmany >> 8) / 7][((pacmanx >> 8) / 7) - 1] > 2)
                        pacmandir = 0;
                break;
                case DIR_DOWN:
                    if(walls[((pacmany >> 8) / 7) + 1][(pacmanx >> 8) / 7] > 2)
                        pacmandir = 0;
                break;
                case DIR_RIGHT:
                    if(walls[(pacmany >> 8) / 7][((pacmanx >> 8) / 7) + 1] > 2)
                        pacmandir = 0;
                break;
            }
        }

        if(pacmandir == DIR_UP)
            pacmany -= pacmanvel;
        if(pacmandir == DIR_LEFT)
            pacmanx -= pacmanvel;
        if(pacmandir == DIR_DOWN)
            pacmany += pacmanvel;
        if(pacmandir == DIR_RIGHT)
            pacmanx += pacmanvel;

        if( ((pacmandir | pacmannextdir) == (DIR_DOWN | DIR_UP)) || ((pacmandir | pacmannextdir) == (DIR_LEFT | DIR_RIGHT)))//if opposite directions change direction
            pacmandir = pacmannextdir;
        else if(!(((pacmanx >> 8) - 3) % 7) && !(((pacmany >> 8) - 3) % 7)){
            switch(pacmannextdir){
                case DIR_UP:
                    if(walls[((pacmany >> 8) / 7) - 1][(pacmanx >> 8) / 7] < 3)
                        pacmandir = pacmannextdir;
                break;
                case DIR_LEFT:
                    if(walls[(pacmany >> 8) / 7][((pacmanx >> 8) / 7) - 1] < 3)
                        pacmandir = pacmannextdir;
                break;
                case DIR_DOWN:
                    if(walls[((pacmany >> 8) / 7) + 1][(pacmanx >> 8) / 7] < 3)
                        pacmandir = pacmannextdir;
                break;
                case DIR_RIGHT:
                    if(walls[(pacmany >> 8) / 7][((pacmanx >> 8) / 7) + 1] < 3)
                        pacmandir = pacmannextdir;
                break;
            }

            //else{//if 90 degree change go diagonally (or if same direction as before do not change anything)
            //    pacmandir |= pacmannextdir;
            //}
        }

        for(i=0;i<4;i++){
            if( !(((enemyx[i] >> 8) - 3) % 7) && enemylock[i] != ((enemyx[i]>>8)/7) + ((enemyy[i]>>8)/7) && !(((enemyy[i] >> 8) - 3) % 7) ){
                enemydir[i] = enemynextdir[i];
                enemynextdir[i] = newdir(enemydir[i], (enemyx[i]>>8)/7, (enemyy[i]>>8)/7, (pacmanx>>8)/7, (pacmany>>8)/7);
                enemylock[i] = ((enemyx[i]>>8)/7) + ((enemyy[i]>>8)/7);//it mustn't change its direction while still at the same place
            }
        }
        if(pacmanx>>8 <= 4)
            pacmanx = (28*7-6)<<8;
        else if(pacmanx>>8 >= 28*7 - 5)
            pacmanx = 5<<8;
        for(i=0;i<4;i++){
            if(enemyx[i]>>8 <= 4)
                enemyx[i] = (28*7 - 6)<<8;
            else if(enemyx[i]>>8 >= 28*7 - 5)
                enemyx[i] = 5<<8;
        }
        for(i=0;i<4;i++){
            if((pacmanx>>8)/7 == (enemyx[i]>>8)/7){
                if((pacmany>>8)/7 == (enemyy[i]>>8)/7){
                    if(mode[i] < 2){
                        if(lifes == 0)
                            return 0;
                        lifes--;
                        pacmanx = (14*7) << 8;
                        pacmany = (23*7 + 3) << 8;
                        pacmandir = DIR_LEFT;
                        sleep(1000);
                    }
                    else{
                        enemyx[i] = (14*7)<<8;
                        enemyy[i] = (11*7+3)<<8;
                        enemydir[i] = DIR_LEFT;
                        enemynextdir[i] = DIR_LEFT;
                        score += 200;
                        mode[i] = 0;
                    }
                }
            }
        }
        if(frightenedtimeelapsed(0) > frighttime)
            memset(mode,0,4);
    }
    free(scrbuf);
    return 0;
}
